/*
 * This file is part of Deliantra, the Roguelike Realtime MMORPG.
 * 
 * Copyright (©) 2005,2006,2007,2008 Marc Alexander Lehmann / Robin Redeker / the Deliantra team
 * Copyright (©) 2001,2007 Mark Wedel & Crossfire Development Team
 * Copyright (©) 1992,2007 Frank Tore Johansen
 * 
 * Deliantra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * The authors can be reached via e-mail to <support@deliantra.net>
 */

#ifndef OBJECT_H
#define OBJECT_H

#include <bitset>

#include "cfperl.h"
#include "shstr.h"

typedef int tag_t;

enum {
  body_skill,
  body_combat,
  body_range,
  body_shield,
  body_arm,
  body_torso,
  body_head,
  body_neck,
  body_finger,
  body_shoulder,
  body_foot,
  body_hand,
  body_wrist,
  body_waist,
  NUM_BODY_LOCATIONS
};

enum slottype_t
{
  slot_none,
  slot_combat,
  slot_ranged,
};

/* See common/item.c */

typedef struct Body_Locations
{
  keyword     save_name;	/* Name used to load/save it to disk */
  const char *use_name;		/* Name used when describing an item we can use */
  const char *nonuse_name;	/* Name to describe objects we can't use */
} Body_Locations;

extern Body_Locations body_locations[NUM_BODY_LOCATIONS];

#define NUM_COINS 4             /* number of coin types */
extern const char *const coins[NUM_COINS + 1];

/*
 * Each object (this also means archetypes!) could have a few of these
 * "dangling" from it; this could also end up containing 'parse errors'.
 *
 * key and value are shared-strings.
 *
 * Please use kv_get/kv_set/kv_del from object rather than
 * accessing the list directly.
 * Exception is if you want to walk this list for some reason.
 */
struct key_value : slice_allocated
{
  key_value *next;
  shstr key, value;
};

struct UUID
{
  uint64 seq;

  static UUID cur; // last uuid generated
  static void init ();
  static UUID gen ();

  UUID () { }
  UUID (uint64 seq) : seq(seq) { }
  operator uint64() { return seq; }
  void operator =(uint64 seq) { this->seq = seq; }

  typedef char BUF [32];

  bool parse (const char *s)
  {
    return sscanf (s, "<1.%" SCNx64 ">", &seq) == 1;
  }

  const char *c_str (char *buf, int len) const
  {
    snprintf (buf, len, "<1.%" PRIx64 ">", seq);

    return buf;
  }

  const char *c_str () const
  {
    static BUF buf;

    return c_str (buf, sizeof (buf));
  }
};

/* Definition for WILL_APPLY values.  Replaces having harcoded values
 * sprinkled in the code.  Note that some of these also replace fields
 * that were in the can_apply area.  What is the point of having both
 * can_apply and will_apply?
 */
#define WILL_APPLY_HANDLE       0x01
#define WILL_APPLY_TREASURE     0x02
#define WILL_APPLY_EARTHWALL    0x04
#define WILL_APPLY_DOOR         0x08
#define WILL_APPLY_FOOD         0x10

struct body_slot
{
  signed char info:4;	/* body info as loaded from the file */
  signed char used:4;	/* Calculated value based on items equipped */
};

INTERFACE_CLASS (object)
// these are being copied
struct object_copy : attachable
{
  typedef bitset<NUM_FLAGS> flags_t;

  sint16 ACC (RW, x), ACC (RW, y);	/* Position in the map for this object */

  uint8 ACC (RW, type);		/* PLAYER, BULLET, etc.  See define.h */
  uint8 ACC (RW, subtype);	/* subtype of object */
  sint8 ACC (RW, direction);	/* Means the object is moving that way. */
  sint8 ACC (RW, facing);	/* Object is oriented/facing that way. */

  shstr ACC (RW, name);		/* The name of the object, obviously... */
  shstr ACC (RW, name_pl);	/* The plural name of the object */
  shstr ACC (RW, title);	/* Of foo, etc */
  shstr ACC (RW, race);		/* human, goblin, dragon, etc */
  shstr ACC (RW, slaying);	/* Which race to do double damage to */
				/* If this is an exit, this is the filename */
  shstr ACC (RW, tag);          // a tag used to tracking this object
  shstr ACC (RW, skill);	/* Name of the skill this object uses/grants */
  shstr ACC (RW, msg);		/* If this is a book/sign/magic mouth/etc */
  shstr ACC (RW, lore);		/* Obscure information about this object, */
				/* To get put into books and the like. */
  shstr ACC (RW, materialname);	/* specific material name */
  shstr ACC (RW, custom_name);	/* Custom name assigned by player */
//  materialtype_t *ACC (RW, material);	/* What material this object consists of */
  object_ptr ACC (RW, owner);	/* Pointer to the object which controls this one */
  object_ptr ACC (RW, enemy);	/* Monster/player to follow even if not closest */
  object_ptr ACC (RW, attacked_by);	/* This object start to attack us! only player & monster */
  object_ptr ACC (RW, chosen_skill);	/* the skill chosen to use */
  object_ptr ACC (RW, spellitem);	/* Spell ability monster is choosing to use */
  object_ptr ACC (RW, spell);	/* Spell that was being cast */
  object_ptr ACC (RW, current_weapon);	/* Pointer to the weapon currently used */
  arch_ptr ACC (RW, arch);	/* Pointer to archetype */
  arch_ptr ACC (RW, other_arch);/* Pointer used for various things - mostly used for what */

  float ACC (RW, speed);	/* The overall speed of this object */
  float ACC (RW, speed_left);	/* How much speed is left to spend this round */
  uint32 ACC (RW, nrof);	/* How many of the objects */

  /* This next big block is basically used for monsters and equipment */
  uint16 ACC (RW, client_type);	/* Public type information.  see doc/Developers/objects */

  sint16 ACC (RW, resist[NROFATTACKS]);	/* Resistance adjustments for attacks */
  uint32 ACC (RW, attacktype);	/* Bitmask of attacks this object does */
  uint32 ACC (RW, path_attuned);/* Paths the object is attuned to */
  uint32 ACC (RW, path_repelled);	/* Paths the object is repelled from */
  uint32 ACC (RW, path_denied);	/* Paths the object is denied access to */
  uint16 ACC (RW, materials);	/* What materials this object consists of */
  sint8 ACC (RW, magic);	/* Any magical bonuses to this item */
  uint8 ACC (RW, state);	/* How the object was last drawn (animation) */
  sint32 ACC (RW, value);	/* How much money it is worth (or contains) */
  /* Note that the last_.. values are sometimes used for non obvious
   * meanings by some objects, eg, sp penalty, permanent exp.
   */
  sint32 ACC (RW, last_heal);	/* Last healed. Depends on constitution */
  sint32 ACC (RW, last_sp);	/* As last_heal, but for spell points */
  sint16 ACC (RW, last_grace);	/* as last_sp, except for grace */
  sint16 ACC (RW, last_eat);	/* How long since we last ate */
  sint16 ACC (RW, invisible);	/* How much longer the object will be invis */
  sint16 ACC (RW, level);	/* Level of creature or object */
  uint8 ACC (RW, pick_up);	/* See crossfire.doc */
  sint8 ACC (RW, item_power);	/* power rating of the object */
  sint8 ACC (RW, gen_sp_armour);/* sp regen penalty this object has (was last_heal) */
  sint8 ACC (RW, glow_radius);	/* indicates the glow radius of the object */
  sint32 ACC (RW, weight);	/* Attributes of the object */
  sint32 ACC (RW, weight_limit);/* Weight-limit of object */
  sint32 ACC (RW, carrying);	/* How much weight this object contains, must be 0 if nrof == 0 */
  sint64 ACC (RW, perm_exp);	/* Permanent exp */
  uint32 ACC (RW, weapontype);	/* type of weapon */
  uint32 ACC (RW, tooltype);	/* type of tool or build facility */
  body_slot slot [NUM_BODY_LOCATIONS];
  faceidx ACC (RW, face);	/* the graphical face */
  faceidx ACC (RW, sound);      /* the sound face */
  faceidx ACC (RW, sound_destroy); /* played on destroy */
  living ACC (RO, stats);	/* Str, Con, Dex, etc */
  /* See the pod/objects.pod for more info about body locations */

  /* Following mostly refers to fields only used for monsters */
  uint32 ACC (RW, hide);	/* The object is hidden, not invisible */

  /* allows different movement patterns for attackers */
  sint32 ACC (RW, move_status);	/* What stage in attack mode */
  uint16 ACC (RW, attack_movement);	/* What kind of attack movement */
  uint16 ACC (RW, run_away);	/* Monster runs away if it's hp goes below this percentage. */
  float ACC (RW, expmul);	/* needed experience = (calc_exp*expmul) - means some */
				/* races/classes can need less/more exp to gain levels */

  /* Spell related information, may be useful elsewhere
   * Note that other fields are used - these files are basically
   * only used in spells.
   */
  sint16 ACC (RW, duration);	/* How long the spell lasts */
  sint16 ACC (RW, casting_time);/* time left before spell goes off */

  uint16 ACC (RW, start_holding);
  uint8 ACC (RW, duration_modifier);	/* how level modifies duration */
  uint8 ACC (RW, dam_modifier);	/* How going up in level effects damage */

  sint8 ACC (RW, range);	/* Range of the spell */
  uint8 ACC (RW, range_modifier);	/* How going up in level effects range  */
  MoveType ACC (RW, move_type);	/* Type of movement this object uses */
  MoveType ACC (RW, move_block);/* What movement types this blocks */

  MoveType ACC (RW, move_allow);/* What movement types explicitly allowd */
  MoveType ACC (RW, move_on);	/* Move types affected moving on to this space */
  MoveType ACC (RW, move_off);	/* Move types affected moving off this space */
  MoveType ACC (RW, move_slow);	/* Movement types this slows down */

  float ACC (RW, move_slow_penalty);	/* How much this slows down the object */

  char *ACC (RW, spellarg);

  /* Following are values used by any object */
  /* this objects turns into or what this object creates */
  treasurelist *ACC (RW, randomitems);	/* Items to be generated */
  flags_t flag;                 /* various flags */
#if FOR_PERL
  bool ACC (RW, flag[NUM_FLAGS]);
#endif
  uint16 ACC (RW, animation_id);/* An index into the animation array */
  uint8 ACC (RW, anim_speed);	/* ticks between animation-frames */
  uint8 ACC (RW, last_anim);	/* last sequence used to draw face */
  sint16 ACC (RW, elevation);	/* elevation of this terrain - not currently used */
  uint8 ACC (RW, smoothlevel);	/* how to smooth this square around */
  uint8 ACC (RW, will_apply);	/* See crossfire.doc */
};

struct object : zero_initialised, object_copy
{
  // These variables are not changed by ->copy_to
  maptile_ptr ACC (RW, map);	/* Pointer to the map in which this object is present */

  UUID ACC (RW, uuid);          // Unique Identifier, survives saves etc.
  int ACC (RO, count);
  object_vector_index ACC (RO, index);	// index into objects
  object_vector_index ACC (RO, active); // index into actives

  player_ptr ACC (RW, contr);	/* Pointer to the player which control this object, ALWAYS set *iff* type == PLAYER */

  object *ACC (RW, below);	/* Pointer to the object stacked below this one */
  object *ACC (RW, above);	/* Pointer to the object stacked above this one */
                                /* Note: stacked in the *same* environment */
  object *inv;			/* Pointer to the first object in the inventory */

  //TODO: container must move into client
  object_ptr ACC (RW, container);	/* Current container being used.  I think this
				 * is only used by the player right now.
				 */
  object *ACC (RW, env);	/* Pointer to the object which is the environment.
				 * This is typically the container that the object is in.
				 */
  object *ACC (RW, more);	/* Pointer to the rest of a large body of objects */
  object *head;			/* Points to the main object of a large body */ // NO ACC, perl semantics are different
  client_container *seen_by;    // seen by which player/container currently?
  key_value *key_values;	/* Fields not explictly known by the loader. */

  // privates / perl
  const shstr &kv_get (const shstr &key) const;
  void kv_del (const shstr &key);
  void kv_set (const shstr &key, const shstr &value);

  // custom extra fields management
  struct key_value_access_proxy
  {
    object &ob;
    shstr key;

    key_value_access_proxy (object &ob, const shstr &key)
    : ob (ob), key (key)
    {
    }

    const key_value_access_proxy &operator =(const shstr &value) const
    {
      ob.kv_set (key, value);
      return *this;
    }

    operator const shstr &() const { return ob.kv_get (key); }
    operator const char  *() const { return ob.kv_get (key); }

  private:
    void operator =(int);
  };

  // operator [] is too annoying to use
  const key_value_access_proxy kv (const shstr &key)
  {
    return key_value_access_proxy (*this, key);
  }

  bool parse_kv (object_thawer &f); // parse kv pairs, (ab-)used by archetypes, which should not exist at all
  MTH void post_load_check ();          // do some adjustments after parsing
  static object *read (object_thawer &f, maptile *map = 0); // map argument due to toal design bogosity, must go.
  bool write (object_freezer &f);

  MTH int slottype () const;
  MTH static object *create ();
  const mapxy &operator =(const mapxy &pos);
  MTH void copy_to (object *dst);
  MTH object *clone (); // create + copy_to a single object
  MTH object *deep_clone (); // copy whole more chain and inventory
  void do_destroy ();
  void gather_callbacks (AV *&callbacks, event_type event) const;
  MTH void destroy (bool destroy_inventory = false);

  // recursively destroy all objects in inventory, optionally dropping them to the ground instead
  MTH void destroy_inv (bool drop_to_ground = false);
  MTH object *insert (object *item); // insert into inventory
  MTH void play_sound (faceidx sound);

  void do_remove ();
  MTH void remove ()
  {
    if (!flag [FLAG_REMOVED])
      do_remove ();
  }

  MTH bool blocked (maptile *m, int x, int y) const;

  void move_to (const mapxy &pos)
  {
    remove ();
    *this = pos;
    insert_at (this, this);
  }

  static bool can_merge_slow (object *op1, object *op2);

  // this is often used in time-critical code, so optimise
  MTH static bool can_merge (object *op1, object *op2)
  {
    return op1->value == op2->value
        && op1->name  == op2->name
        && can_merge_slow (op1, op2);
  }

  MTH void set_owner (object *owner);
  MTH void set_speed (float speed);
  MTH bool change_weapon (object *ob);
  MTH bool change_skill (object *ob);

  MTH void open_container (object *new_container);
  MTH void close_container ()
  {
    open_container (0);
  }

  MTH object *force_find (const shstr name);
  MTH void force_add (const shstr name, int duration = 0);

  // overwrite the attachable should_invoke function with a version that also checks ev_want_type
  bool should_invoke (event_type event)
  {
    return ev_want_event [event] || ev_want_type [type] || cb;
  }

  MTH void instantiate ();

  // recalculate all stats
  MTH void update_stats ();
  MTH void roll_stats ();
  MTH void swap_stats (int a, int b);
  MTH void add_statbonus ();
  MTH void remove_statbonus ();
  MTH void drain_stat ();
  MTH void drain_specific_stat (int deplete_stats);
  MTH void change_luck (int value);

  // info must hold 256 * 3 bytes currently
  const char *debug_desc (char *info) const;
  MTH const char *debug_desc () const; // uses at least 3 round-robin buffers
  const char *flag_desc (char *desc, int len) const;

  MTH bool decrease (sint32 nr = 1); // returns true if anything is left
  MTH object *split (sint32 nr = 1); // return 0 on failure

  MTH int number_of () const
  {
    return nrof ? nrof : 1;
  }

  MTH sint32 total_weight () const
  {
    return (weight + carrying) * number_of ();
  }

  MTH void update_weight ();

  // return the dominant material of this item, always return something
  const materialtype_t *dominant_material () const;

  // return the volume of this object in cm³
  MTH uint64 volume () const
  {
    return total_weight ()
           * 1000
           * (type == CONTAINER ? 1000 : 1)
           / dominant_material ()->density;
  }

  MTH bool is_weapon () const { return type == ARROW || type == BOW || type == WEAPON; }
  MTH bool is_armor  () const { return type == ARMOUR  || type == SHIELD || type == HELMET
                                    || type == CLOAK   || type == BOOTS  || type == GLOVES
                                    || type == BRACERS || type == GIRDLE; }
  MTH bool is_alive  () const { return (type == PLAYER
                                        || flag [FLAG_MONSTER]
                                        || (flag [FLAG_ALIVE] && !flag [FLAG_GENERATOR] && type != DOOR))
                                       && !flag [FLAG_IS_A_TEMPLATE]; }
  MTH bool is_arrow  () const { return type == ARROW
                                       || (type == SPELL_EFFECT
                                           && (subtype == SP_BULLET || subtype == SP_MAGIC_MISSILE)); }
  MTH bool is_range  () const { return type == BOW || type == ROD || type == WAND || type == HORN; }

  MTH bool has_active_speed () const { return FABS(speed) >= MIN_ACTIVE_SPEED; }

  // temporary: wether the object can be saved in a map file
  // contr => is a player
  // head  => only save head of a multitile object
  // owner => can not reference owner yet
  MTH bool can_map_save () const { return !head && (!owner || owner->contr) && !contr && !flag [FLAG_NO_MAP_SAVE]; }

  /* This return true if object has still randomitems which
   * could be expanded.
   */
  MTH bool has_random_items () const { return randomitems && !flag [FLAG_IS_A_TEMPLATE]; }

  // returns the outermost owner, never returns 0
  MTH object *outer_owner ()
  {
    for (object *op = this; ; op = op->owner)
      if (!op->owner)
        return op;
  }

  // returns the outermost environment, never returns 0
  MTH object *outer_env ()
  {
    for (object *op = this; ; op = op->env)
      if (!op->env)
        return op;
  }

  // returns the player that has this object in his inventory, or 0
  // we assume the player is always the outer env
  MTH object *in_player ()
  {
    object *op = outer_env ();

    return op->type == PLAYER ? op : 0;
  }

  // "temporary" helper function
  MTH object *head_ ()
  {
    return head ? head : this;
  }

  MTH bool is_head ()
  {
    return head_ () == this;
  }

  MTH bool is_on_map () const
  {
    return !env && !flag [FLAG_REMOVED];
  }

  // returns the player that cna see this object, if any
  MTH object *visible_to () const;

  MTH std::string long_desc        (object *who = 0);
  MTH std::string describe_monster (object *who = 0);
  MTH std::string describe_item    (object *who = 0);
  MTH std::string describe         (object *who = 0);

  // If this object has no extra parts but should have them,
  // add them, effectively expanding heads into multipart
  // objects. This method only works on objects not inserted
  // anywhere.
  MTH void expand_tail ();

  MTH void create_treasure (treasurelist *tl, int flags = 0);

  // insert object at same map position as 'where'
  // handles both inventory and map "positions"
  MTH object *insert_at (object *where, object *originator = 0, int flags = 0);
  MTH void drop_unpaid_items ();

  MTH void activate   ();
  MTH void deactivate ();
  MTH void activate_recursive   ();
  MTH void deactivate_recursive ();

  // set the given flag on all objects in the inventory recursively
  MTH void set_flag_inv (int flag, int value = 1);

  void enter_exit (object *exit);//Perl
  MTH void enter_map (maptile *newmap, int x, int y);
  void player_goto (const char *path, int x, int y); // only for players

  // returns the mapspace this object is in
  mapspace &ms () const;

  // fully recursive iterator
  struct iterator_base
  {
    object *item;

    iterator_base (object *container)
    : item (container)
    {
    }

    operator object *() const { return item; }

    object *operator ->() const { return  item; }
    object &operator * () const { return *item; }
  };

  MTH unsigned int random_seed () const
  {
    return (unsigned int)uuid.seq;
  }

  // depth-first recursive iterator
  struct depth_iterator : iterator_base
  {
    depth_iterator (object *container);
    void next ();
    object *operator ++(   ) { next (); return item; }
    object *operator ++(int) { object *i = item; next (); return i; }
  };

  object *begin ()
  {
    return this;
  }

  object *end ()
  {
    return this;
  }

  /* This returns TRUE if the object is something that
   * a client might want to know about.
   */
  MTH bool client_visible () const
  {
    return !invisible && type != PLAYER;
  }

  // the client does nrof * this weight
  MTH sint32 client_weight () const
  {
    return weight + carrying;
  }

  MTH struct region *region () const;

  void statusmsg (const char *msg, int color = NDI_BLACK);
  void failmsg (const char *msg, int color = NDI_RED);

  const char *query_inventory (object *who = 0, const char *indent = "");

  MTH const_octet_string ref () const; // creates and returns a consistent persistent object reference
  static object *deref (const_octet_string ref); // returns the object from the generated refreence, if possible

protected:
  void link   ();
  void unlink ();

  object ();
  ~object ();

private:
  object &operator =(const object &);
  object (const object &);
};

// move this object to the top of its env's inventory to speed up
// searches for it.
static object *
splay (object *ob)
{
  if (ob->env && ob->env->inv != ob)
    {
      if (ob->above) ob->above->below = ob->below;
      if (ob->below) ob->below->above = ob->above;

      ob->above = 0;
      ob->below = ob->env->inv;
      ob->below->above = ob;
      ob->env->inv = ob;
    }

  return ob;
}

typedef struct oblnk
{				/* Used to link together several objects */
  object_ptr ob;
  struct oblnk *next;
} objectlink;

typedef struct oblinkpt
{				/* Used to link together several object links */
  struct oblnk *link;
  sint32 value;			/* Used as connected value in buttons/gates */
  struct oblinkpt *next;
} oblinkpt;

object *find_skill_by_name (object *who, const char *name);
object *find_skill_by_name (object *who, const shstr &sh);
object *find_skill_by_number (object *who, int skillno);

/*
 * The archetype structure is a set of rules on how to generate and manipulate
 * objects which point to archetypes.
 * This probably belongs in arch.h, but there really doesn't appear to
 * be much left in the archetype - all it really is is a holder for the
 * object and pointers.  This structure should get removed, and just replaced
 * by the object structure
 */

INTERFACE_CLASS (archetype)
struct archetype : object
{
  static arch_ptr empty; // the empty_archetype
  MTH static void gc ();

  archetype (const char *name);
  ~archetype ();
  void gather_callbacks (AV *&callbacks, event_type event) const;

  MTH static archetype *find (const_utf8_string name);

  MTH void link   ();
  MTH void unlink ();

  MTH static object *get (const char *name); // (find() || singularity)->instance()
  MTH object *instance ();

  object_vector_index ACC (RW, archid);         // index in archvector
  shstr ACC (RW, archname);		   	/* More definite name, like "generate_kobold" */

  sint8 ACC (RW, min_x), ACC (RW, min_y);       /* extents, compared to the head (min_x, min_y should be zero, but aren't...) */
  sint8 ACC (RW, max_x), ACC (RW, max_y);       

  // support for archetype loading
  static archetype *read (object_thawer &f);
  MTH static void commit_load (); // commit any objects loaded, resolves cyclic dependencies and more
  static void postpone_arch_ref (arch_ptr &ref, const_utf8_string other_arch); /* postpone  other_arch reference */
};

// compatbiility, remove once replaced by ->instance
inline object *
arch_to_object (archetype *at)
{
  return at->instance ();
}

inline void
object_freezer::put (keyword k, archetype *v)
{
  put (k, v ? &v->archname : (const char *)0);
}

typedef object_vector<object,    &object::index >    objectvec;
typedef object_vector<object,    &object::active>    activevec;
typedef object_vector<archetype, &archetype::archid> archvec;

extern objectvec objects;
extern activevec actives;
extern archvec   archetypes;

// "safely" iterate over inv in a way such that the current item is removable
// quite horrible, thats why its hidden in some macro
#define for_inv_removable(op,var)			\
  for (object *var, *next_ = (op)->inv; (var = next_), var && (next_ = var->below), var; )

#define for_all_objects(var)                    	\
  for (unsigned _i = 0; _i < objects.size (); ++_i)     \
    statementvar (object *, var, objects [_i])

#define for_all_actives(var)                    	\
  for (unsigned _i = 0; _i < actives.size (); ++_i)     \
    statementvar (object *, var, actives [_i])

#define for_all_archetypes(var)                 	\
  for (unsigned _i = 0; _i < archetypes.size (); ++_i)  \
    statementvar (archetype *, var, archetypes [_i])

/* Used by update_object to know if the object being passed is
 * being added or removed.
 */
#define UP_OBJ_INSERT   1
#define UP_OBJ_REMOVE   2
#define UP_OBJ_CHANGE   3
#define UP_OBJ_FACE     4	/* Only thing that changed was the face */

/* These are flags passed to insert_ob_in_map and 
 * insert_ob_in_ob.  Note that all flags may not be meaningful
 * for both functions.
 * Most are fairly explanatory:
 * INS_NO_MERGE: don't try to merge inserted object with ones alrady
 *    on space.
 * INS_ABOVE_FLOOR_ONLY: Put object immediatly above the floor.
 * INS_NO_WALK_ON: Don't call check_walk_on against the
 *    originator - saves cpu time if you know the inserted object
 *    is not meaningful in terms of having an effect.
 * INS_ON_TOP: Always put object on top.  Generally only needed when loading
 *     files from disk and ordering needs to be preserved. 
 * INS_BELOW_ORIGINATOR: Insert new object immediately below originator -
 *     Use for treasure chests so the new object is the highest thing
 *     beneath the player, but not actually above it.  Note - the
 *     map and x,y coordinates for the object to be inserted must
 *     match the originator.
 *
 * Note that INS_BELOW_ORIGINATOR, INS_ON_TOP, INS_ABOVE_FLOOR_ONLY
 * are mutually exclusive.  The behaviour for passing more than one
 * should be considered undefined - while you may notice what happens
 * right now if you pass more than one, that could very well change
 * in future revisions of the code.
 */
#define INS_NO_MERGE            0x0001
#define INS_ABOVE_FLOOR_ONLY    0x0002
#define INS_NO_WALK_ON          0x0004
#define INS_ON_TOP              0x0008
#define INS_BELOW_ORIGINATOR    0x0010

#define ARCH_DEPLETION          "depletion"

#endif

